#!/usr/bin/perl
# 
# Starter sample for fallback fonts
#
# Required software: PDFlib/PDFlib+PDI/PPS 9
# Required data: suitable fonts, Japanese CMaps

use PDFlib::PDFlib 9.1;
use strict;

# This is where the data files are. Adjust as necessary. 
use constant searchpath => "../data";
use constant outfile => "starter_fallback.pdf";

use constant {
    llx => 30.0,
    lly => 50.0,
    urx => 800.0,
    ury => 570.0
};

use constant headers => (
    "Use case",
    "Option list for the 'fallbackfonts' option",
    "Base font",
    "With fallback font"
);

# Key names used to make a dictionary for the description of the
# testcase entries
my @testcase_keys = qw(usecase fontname encoding fallbackoptions text);

# Function to create a hash describing each testcase
sub make_testcase_hash {
    my $values = $_[0]; # reference to array
    my %result;
    @result{@testcase_keys} = @{$values};
    return \%result;
}

# The testcases organized as an array of references to hashes
my @testcases = map { make_testcase_hash($_) } (
    [ # Add Euro glyph to an encoding which doesn't support it
      "Extend 8-bit encoding",  # usecase
      "Helvetica",              # fontname
      "iso8859-1",              # encoding
      "{fontname=Helvetica encoding=unicode forcechars=euro}", #fallbackoptions
      # Reference Euro glyph by name (since it is missing from the encoding) 
      "123&euro;"               # text
    ],
    [
      "Use Euro glyph from another font",
      "Courier",
      "unicode",
      "{fontname=Helvetica encoding=unicode forcechars=euro textrise=-5%}",
      "123&euro;"
    ],
    [
      "Add enlarged pictogram",
      "NotoSerif-Regular",
      "unicode",
      # U+261E = WHITE RIGHT POINTING INDEX
      "{fontname=ZapfDingbats encoding=unicode forcechars=U+261E fontsize=150% " .
      "textrise=-15%}",
      "hand symbol: &#x261E;"
    ],
    [
      "Add enlarged symbol glyph",
      "NotoSerif-Regular",
      "unicode",
      "{fontname=Symbol encoding=unicode forcechars=U+2663 fontsize=125%}",
      "club symbol: &#x2663;"
    ],
    [ # Hebrew characters missing in the font will be pulled from Hebrew
      "Add Hebrew characters to Latin font",
      "NotoSerif-Regular",
      "unicode",
      "{fontname=NotoSerifHebrew-Regular encoding=unicode}", 
      "Hebrew: &#x05D0;"
    ],
    [ # Font with end-user defined character (EUDC) 
      "Gaiji with EUDC font",
      "KozMinProVI-Regular",
      "unicode",
      "{fontname=EUDC encoding=unicode forcechars=U+E000 fontsize=140% " .
      "textrise=-20%}",
      "Gaiji: &#xE000;"
    ],
    [ "Replace Latin characters in CJK font",
      "KozMinProVI-Regular",
      "unicode",
      "{fontname=Courier encoding=unicode forcechars={U+0020-U+007E}}",
      "Latin and &#x65E5;&#x672C;&#x8A9E;"
    ],
);

my $p = new PDFlib::PDFlib;

eval {
    my $optlist;
    
    $p->set_option("SearchPath={{" . searchpath ."}}");
    # all strings are expected as utf8
    $p->set_option("stringformat=utf8");
    $p->set_option("charref=true");
    $p->set_option("glyphcheck=replace");

    # This means that formatting and other errors will raise an
    # exception. This simplifies our sample code, but is not
    # recommended for production code.
    $p->set_option("errorpolicy=exception");

    # Set an output path according to the name of the topic 
    if ($p->begin_document(outfile, "") == -1) {
        printf("Error: %s\n", $p->get_errmsg());
        $p->delete();
        exit(2);
    }

    $p->set_info("Creator", "PDFlib starter sample");
    $p->set_info("Title", "starter_fallback");

    # Start Page 
    $p->begin_page_ext(0, 0, "width=a4.height height=a4.width");

    my $table = -1;

    # Table header 
    my $col = 1;
    foreach my $header (headers) {
        $optlist =
           "fittextline={fontname=NotoSerif-Regular encoding=unicode fontsize=10} " .
           "margin=4";
        $table = $p->add_table_cell($table, $col, 1, $header, $optlist);
        $col += 1;
    }

    # Create fallback samples, one use case per row
    my $row = 2;
    foreach my $testcase (@testcases) {
        $col = 1;

        # Column 1: description of the use case 
        $optlist = 
            "fittextline={fontname=NotoSerif-Regular encoding=unicode fontsize=10} " .
            "margin=4";
        $table = $p->add_table_cell($table, $col++, $row,
                                    $testcase->{usecase}, $optlist);

        # Column 2: reproduce option list literally 
        $optlist = 
            "fittextline={fontname=NotoSerif-Regular encoding=unicode fontsize=10} " .
            "margin=4";
        $table = $p->add_table_cell($table, $col++, $row,
                                    $testcase->{fallbackoptions}, $optlist);

        # Column 3: text with base font 
        $optlist = sprintf
             "fittextline={fontname=%s encoding=%s fontsize=10 " .
             "replacementchar=? } margin=4",
             $testcase->{fontname}, $testcase->{encoding};
        $table = $p->add_table_cell($table, $col++, $row,
                                    $testcase->{text}, $optlist);

        # Column 4: text with base font and fallback fonts 
        $optlist = sprintf
             "fittextline={fontname=%s encoding=%s " .
             "fontsize=10 fallbackfonts={%s}} margin=4",
             $testcase->{fontname},
             $testcase->{encoding},
             $testcase->{fallbackoptions};
        $table = $p->add_table_cell($table, $col++, $row,
                                    $testcase->{text}, $optlist);
        
        $row += 1;
    }

    # Place the table 
    $optlist = "header=1 fill={{area=rowodd fillcolor={gray 0.9}}} " .
                "stroke={{line=other}} ";
    my $result = $p->fit_table($table, llx, lly, urx, ury, $optlist);

    if ($result eq "_error")
    {
        printf("Couldn't place table: %s\n", $p->get_errmsg());
        $p->delete();
        exit(2);
    }

    $p->end_page_ext("");
    $p->end_document("");
};

if ($@) {
    die("$0: PDFlib Exception occurred:\n$@");
}
